from .citizen_fields_mixin import CitizenFieldsMixin
from .identity_fields_mixin import IdentityFieldsMixin, IdentityFieldsMixinError
from .personal_fields_mixin import FullNamePersonalFieldsMixin, PersonalFieldsMixin
from .review_fields_mixin import ReviewFieldsMixin
from .sample_collection_fields_mixin import SampleCollectionFieldsMixin
from .scored_review_fields_mixin import ScoredReviewFieldsMixin
from .site_fields_mixin import SiteFieldsMixin
from .verification_fields_mixin import VerificationFieldsMixin
from .vulnerability_fields_mixin import VulnerabilityFieldsMixin



from django.db import models
from django_crypto_fields.fields import IdentityField
from django_crypto_fields.models import CryptoMixin

from ..choices import IDENTITY_TYPE


class IdentityFieldsMixinError(Exception):
    pass


class IdentityFieldsMixin(CryptoMixin, models.Model):

    """
    Note: specify identifier_type CHOICES on the form.
    For example:

        from edc_constants.choices import IDENTITY_TYPE

        identity_type = forms.CharField(
            label='What type of identity number is this?',
            widget=forms.RadioSelect(choices=list(IDENTITY_TYPE)))
    """

    identity = IdentityField(verbose_name="Identity number")

    identity_type = models.CharField(
        verbose_name="What type of identity number is this?",
        max_length=25,
        choices=IDENTITY_TYPE,
    )

    confirm_identity = IdentityField(
        help_text="Retype the identity number", null=True, blank=False
    )

    def save(self, *args, **kwargs):
        if self.identity != self.confirm_identity:
            raise IdentityFieldsMixinError(
                "'Identity' must match 'confirm_identity'. " "Catch this error on the form"
            )
        super().save(*args, **kwargs)

    class Meta:
        abstract = True



from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist

from .exceptions import ConsentVersionSequenceError
from .site_consents import site_consents


class ConsentHelper:

    """A class to get the consent configuration object and to
    validate version numbers if this consent is an update
    of a previous.
    """

    def __init__(
        self,
        model_cls=None,
        update_previous=None,
        subject_identifier=None,
        identity=None,
        first_name=None,
        consent_datetime=None,
        dob=None,
        last_name=None,
        subject_identifier_as_pk=None,
        subject_identifier_aka=None,
        **kwargs,
    ):
        self._previous_consent = None
        self.model_cls = model_cls
        self.dob = dob
        self.first_name = first_name
        self.identity = identity
        self.last_name = last_name
        self.subject_identifier = subject_identifier
        self.update_previous = True if update_previous is None else update_previous

        self.consent_object = site_consents.get_consent_for_period(
            model=self.model_cls._meta.label_lower,
            consent_group=self.model_cls._meta.consent_group,
            report_datetime=consent_datetime,
        )

        # these to be set on the model
        self.version = self.consent_object.version
        self.updates_versions = self.consent_object.updates_versions

        # if updates a previous, validate version sequence
        # and update the subject_identifier_as_pk, etc
        if self.updates_versions and update_previous:
            self.previous_consent.subject_identifier_as_pk = subject_identifier_as_pk
            self.previous_consent.subject_identifier_aka = subject_identifier_aka
            self.previous_consent.save(
                update_fields=["subject_identifier_as_pk", "subject_identifier_aka"]
            )

    @property
    def previous_consent(self):
        """Returns the previous consent or raises if does
        not exist or is out of sequence with the current.
        """
        if not self._previous_consent:
            opts = dict(
                subject_identifier=self.subject_identifier,
                identity=self.identity,
                version__in=self.consent_object.updates_versions,
            )
            opts = {k: v for k, v in opts.items() if v is not None}
            try:
                self._previous_consent = self.model_cls.objects.get(**opts)
            except ObjectDoesNotExist:
                updates_versions = ", ".join(self.consent_object.updates_versions)
                raise ConsentVersionSequenceError(
                    f"Failed to update previous version. A previous consent "
                    f"with version in {updates_versions} for {self.subject_identifier} "
                    f"was not found. Consent version '{self.version}' is "
                    "configured to update a previous version."
                )
            except MultipleObjectsReturned:
                previous_consents = self.model_cls.objects.filter(**opts).order_by("-version")
                self._previous_consent = previous_consents[0]
        return self._previous_consent



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any

from dateutil.relativedelta import relativedelta
from django import forms
from edc_constants.constants import NO, YES
from edc_model_form.utils import get_field_or_raise
from edc_utils import AgeValueError, age, formatted_age

from ...site_consents import ConsentObjectDoesNotExist, SiteConsentError, site_consents
from ...utils import InvalidInitials, verify_initials_against_full_name

if TYPE_CHECKING:
    from ...consent import Consent


class CustomValidationMixin:
    """Form for models that are a subclass of BaseConsent."""

    @property
    def consent_config(self) -> Consent:
        """Returns a consent_config instance or raises on
        missing data.
        """
        try:
            consent_config = site_consents.get_consent(
                report_datetime=self.consent_datetime,
                consent_model=self._meta.model._meta.label_lower,
                consent_group=self._meta.model._meta.consent_group,
            )
        except (ConsentObjectDoesNotExist, SiteConsentError) as e:
            raise forms.ValidationError(e)
        if not consent_config.version:
            raise forms.ValidationError("Unable to determine consent version")
        return consent_config

    def get_field_or_raise(self, name: str, msg: str) -> Any:
        return get_field_or_raise(
            name, msg, cleaned_data=self.cleaned_data, instance=self.instance
        )

    @property
    def consent_datetime(self) -> datetime:
        return self.get_field_or_raise("consent_datetime", "Consent date and time is required")

    @property
    def identity(self) -> str:
        return self.get_field_or_raise("identity", "Identity is required")

    @property
    def confirm_identity(self) -> str:
        return self.get_field_or_raise("confirm_identity", "Confirmed identity is required")

    @property
    def age_delta(self) -> relativedelta | None:
        dob = self.cleaned_data.get("dob")
        if self.consent_datetime and dob:
            try:
                return age(dob, self.consent_datetime)
            except AgeValueError as e:
                raise forms.ValidationError(str(e))
        return None

    def validate_min_age(self) -> None:
        """Raises if age is below the age of consent"""
        if self.age_delta:
            if self.age_delta.years < self.consent_config.age_min:
                raise forms.ValidationError(
                    {
                        "dob": (
                            f"Subject's age is {self.age_delta.years}. "
                            "Subject is not eligible for consent. Minimum age of consent is "
                            f"{self.consent_config.age_min}."
                        )
                    }
                )

    def validate_max_age(self) -> None:
        """Raises if age is above the age of consent"""
        if self.age_delta:
            if self.age_delta.years > self.consent_config.age_max:
                raise forms.ValidationError(
                    {
                        "dob": (
                            f"Subject's age is {self.age_delta.years}. "
                            "Subject is not eligible for consent. Maximum age of consent is "
                            f"{self.consent_config.age_max}."
                        )
                    }
                )

    def validate_identity_and_confirm_identity(self) -> None:
        if self.identity and self.confirm_identity:
            if self.identity != self.confirm_identity:
                msg = (
                    "Identity mismatch. Identity must match "
                    f"the confirmation field. Got {self.identity} != "
                    f"{self.confirm_identity}"
                )
                if "identity" in self.cleaned_data:
                    raise forms.ValidationError({"identity": msg})
                else:
                    raise forms.ValidationError({"__all__": msg})

    def validate_identity_plus_version_is_unique(self) -> None:
        """Enforce a unique constraint on personal identity number
        + consent version.

        Note: since version is not part of cleaned data, django form
        will not do the integrity check by default.
        """
        exclude_opts = dict(id=self.instance.id) if self.instance.id else {}
        if (
            subject_consent := self._meta.model.objects.filter(
                identity=self.identity, version=self.consent_config.version
            )
            .exclude(**exclude_opts)
            .last()
        ):
            msg = (
                "Identity number already submitted for consent "
                f"{self.consent_config.version}. "
                f"See `{subject_consent.subject_identifier}`."
            )
            if "identity" in self.cleaned_data:
                raise forms.ValidationError({"identity": msg})
            else:
                raise forms.ValidationError({"__all__": msg})

    def validate_identity_with_unique_fields(self) -> None:
        cleaned_data = self.cleaned_data
        msg_word = None
        first_name = cleaned_data.get("first_name")
        initials = cleaned_data.get("initials")
        familiar_name = cleaned_data.get("familiar_name")
        dob = cleaned_data.get("dob")
        opts = dict(
            initials=initials,
            dob=dob,
            version=self.consent_config.version,
        )
        if familiar_name:
            opts.update(familiar_name=familiar_name)
            msg_word = "familiar name"
        elif first_name:
            opts.update(first_name=first_name)
            msg_word = "first name"
        if (
            subject_consent := self._meta.model.objects.filter(**opts)
            .exclude(identity=self.identity)
            .last()
        ):
            raise forms.ValidationError(
                f"These personal details ({msg_word}, initials, dob) describe "
                f"another subject. See {subject_consent.subject_identifier} (1)."
            )

    def validate_initials_with_full_name(self) -> None:
        cleaned_data = self.cleaned_data
        try:
            verify_initials_against_full_name(**cleaned_data)
        except InvalidInitials as e:
            raise forms.ValidationError({"initials": str(e)})

    def validate_gender_of_consent(self: Any) -> str:
        """Validates gender is a gender of consent."""
        gender = self.cleaned_data.get("gender")
        if gender not in self.consent_config.gender:
            raise forms.ValidationError(
                "Gender of consent can only be '%(gender_of_consent)s'. Got '%(gender)s'.",
                params={
                    "gender_of_consent": "' or '".join(self.consent_config.gender),
                    "gender": gender,
                },
                code="invalid",
            )
        return gender

    def validate_guardian_and_dob(self) -> None:
        """Validates guardian is required if age is below age_is_adult
        from consent config.
        """
        cleaned_data = self.cleaned_data
        guardian = cleaned_data.get("guardian_name")
        dob = cleaned_data.get("dob")
        rdelta = relativedelta(self.consent_datetime.date(), dob)
        if rdelta.years < self.consent_config.age_is_adult:
            if not guardian:
                raise forms.ValidationError(
                    {
                        "guardian_name": (
                            f"Subject's age is {formatted_age(dob, self.consent_datetime)}. "
                            "Subject is a minor. Guardian's "
                            "name is required with signature on the paper "
                            "document."
                        )
                    }
                )
        if rdelta.years >= self.consent_config.age_is_adult and guardian:
            if guardian:
                raise forms.ValidationError(
                    {
                        "guardian_name": (
                            f"Subject's age is {formatted_age(dob, self.consent_datetime)}. "
                            "Subject is an adult. Guardian's name is NOT required."
                        )
                    }
                )

    def validate_dob_relative_to_consent_datetime(self) -> None:
        """Validates that the dob is within the bounds of MIN and
        MAX set on the model.
        """
        self.validate_min_age()
        self.validate_max_age()

    def validate_is_literate_and_witness(self) -> None:
        cleaned_data = self.cleaned_data
        is_literate = cleaned_data.get("is_literate")
        witness_name = cleaned_data.get("witness_name")
        if is_literate == NO and not witness_name:
            raise forms.ValidationError(
                {
                    "witness_name": "Provide a name of a witness on this form and "
                    "ensure paper consent is signed."
                }
            )
        if is_literate == YES and witness_name:
            raise forms.ValidationError({"witness_name": "This field is not required"})



from __future__ import annotations

from datetime import datetime

from django.apps import apps as django_apps
from edc_constants.constants import FEMALE, MALE


class InvalidGender(Exception):
    pass


class NaiveDatetimeError(Exception):
    pass


class Consent:
    default_version = "1"
    default_subject_type = "subject"
    default_consent_group = django_apps.get_app_config("edc_consent").default_consent_group

    def __init__(
        self,
        model: str,
        group: str = None,
        start: datetime = None,
        end: datetime = None,
        version: str | None = None,
        gender: list[str] = None,
        updates_versions: list[str] | str | None = None,
        age_min: int = None,
        age_max: int = None,
        age_is_adult: int | None = None,
        subject_type: str | None = None,
        proxy_models: list[str] | None = None,
    ):
        """A class that represents the general attributes
        of a consent.
        """
        if not start.tzinfo:
            raise NaiveDatetimeError(f"Naive datetime is invalid. Got {start}.")
        if not end.tzinfo:
            raise NaiveDatetimeError(f"Naive datetime is invalid. Got {end}.")
        if MALE not in gender and FEMALE not in gender:
            raise InvalidGender(f"Invalid gender. Got {gender}.")
        self.model = model
        self.proxy_models = proxy_models or []
        self.group = group or self.default_consent_group
        self.start = start
        self.end = end
        self.updates_versions = updates_versions or []
        self.version = version or self.default_version
        self.gender = gender
        self.age_min = age_min
        self.age_max = age_max
        self.age_is_adult = age_is_adult
        self.subject_type = subject_type or self.default_subject_type
        self.verbose_name = f"{self.model} {self.version}"
        self.name = f"{self.model}-{self.version}"
        if self.updates_versions:
            if not isinstance(self.updates_versions, (list, tuple)):
                self.updates_versions = [
                    x.strip() for x in self.updates_versions.split(",") if x.strip() != ""
                ]

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}({self.model}, {self.version}) "
            f"from {self.start} to {self.end}>"
        )

    def __str__(self):
        return self.verbose_name

    @property
    def model_cls(self):
        return django_apps.get_model(self.model)



from __future__ import annotations

from datetime import date, datetime
from typing import TYPE_CHECKING

from dateutil.relativedelta import relativedelta
from django.conf import settings
from django.contrib.sites.models import Site
from edc_constants.constants import FEMALE
from edc_utils import get_utcnow, get_utcnow_as_date
from model_bakery import baker

from edc_consent.utils import get_consent_model_name

if TYPE_CHECKING:
    from edc_screening.tests.models import SubjectScreening


class ConsentTestCaseMixin:
    @staticmethod
    def get_subject_consent(
        subject_screening: SubjectScreening | None = None,
        consent_datetime: datetime | None = None,
        site_id: int | None = None,
        dob: date | None = None,
        screening_identifier: str | None = None,
        gender: str | None = None,
        initials: str | None = None,
        age_in_years: int | None = None,
        report_datetime: datetime | None = None,
    ):
        now = get_utcnow()
        now_as_date = get_utcnow_as_date()
        screening_identifier = (
            subject_screening.screening_identifier
            if subject_screening
            else (screening_identifier or "ABCDEF")
        )
        initials = subject_screening.initials if subject_screening else (initials or "XX")
        gender = subject_screening.gender if subject_screening else (gender or FEMALE)
        age_in_years = (
            subject_screening.age_in_years if subject_screening else (age_in_years or 25)
        )
        dob = dob or (now_as_date - relativedelta(years=age_in_years))
        report_datetime = (
            subject_screening.report_datetime if subject_screening else report_datetime or now
        )
        consent_datetime = consent_datetime or report_datetime

        return baker.make_recipe(
            get_consent_model_name(),
            user_created="erikvw",
            user_modified="erikvw",
            screening_identifier=screening_identifier,
            initials=initials,
            gender=gender,
            dob=dob,
            site=Site.objects.get(id=site_id or settings.SITE_ID),
            consent_datetime=consent_datetime,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from django import forms
from django.core.exceptions import ObjectDoesNotExist
from edc_utils import floor_secs, formatted_date, formatted_datetime
from edc_utils.date import to_local, to_utc

from ..constants import DEFAULT_CONSENT_GROUP
from ..site_consents import site_consents
from ..utils import get_consent_model_name

if TYPE_CHECKING:
    from ..model_mixins import ConsentModelMixin


class RequiresConsentModelFormMixin:

    """Model form mixin for CRF or PRN forms to access the consent.

    Use with CrfModelMixin, etc
    """

    def clean(self):
        cleaned_data = super().clean()
        self.validate_against_consent()
        return cleaned_data

    def validate_against_consent(self) -> None:
        """Raise an exception if the report datetime doesn't make
        sense relative to the consent.
        """
        if self.report_datetime:
            consent = self.get_consent_or_raise()
            if floor_secs(to_utc(self.report_datetime)) < floor_secs(consent.consent_datetime):
                dte_str = formatted_datetime(to_local(consent.consent_datetime))
                raise forms.ValidationError(
                    f"Report datetime cannot be before consent datetime. Got {dte_str}."
                )
            if to_utc(self.report_datetime).date() < consent.dob:
                dte_str = formatted_date(consent.dob)
                raise forms.ValidationError(
                    f"Report datetime cannot be before DOB. Got {dte_str}"
                )

    @property
    def consent_group(self) -> str:
        try:
            consent_group = self._meta.model._meta.consent_group
        except AttributeError:
            consent_group = DEFAULT_CONSENT_GROUP
        return consent_group

    @property
    def consent_model(self) -> str:
        return get_consent_model_name()

    def get_consent_or_raise(self) -> ConsentModelMixin:
        """Return an instance of the consent model"""
        consent_object = site_consents.get_consent(
            report_datetime=self.report_datetime,
            consent_group=self.consent_group,
            consent_model=self.consent_model,
        )
        try:
            obj = consent_object.model_cls.consent.consent_for_period(
                subject_identifier=self.subject_identifier,
                report_datetime=self.report_datetime,
            )
        except ObjectDoesNotExist:
            raise forms.ValidationError(
                f"`{consent_object.model_cls._meta.verbose_name}` does not exist "
                f"to cover this subject on {formatted_datetime(self.report_datetime)}"
            )
        return obj



from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from edc_identifier import is_subject_identifier_or_raise
from edc_model_wrapper import ModelWrapper
from edc_utils import get_uuid

from ..site_consents import site_consents


class ConsentModelWrapperMixin(ModelWrapper):
    consent_model_wrapper_cls = None

    @property
    def consent_object(self):
        """Returns a consent configuration object from site_consents
        relative to the wrapper's "object" report_datetime.
        """
        default_consent_group = django_apps.get_app_config("edc_consent").default_consent_group
        consent_object = site_consents.get_consent_for_period(
            model=self.consent_model_wrapper_cls.model,
            report_datetime=self.object.report_datetime,
            consent_group=default_consent_group,
        )
        return consent_object

    @property
    def consent_model_obj(self):
        """Returns a consent model instance or None."""
        consent_model_cls = django_apps.get_model(self.consent_model_wrapper_cls.model)
        try:
            model_obj = consent_model_cls.objects.get(**self.consent_options)
        except ObjectDoesNotExist:
            model_obj = None
        return model_obj

    @property
    def consent(self):
        """Returns a wrapped saved or unsaved consent."""
        model_obj = self.consent_model_obj or self.consent_object.model_cls(
            **self.create_consent_options
        )
        return self.consent_model_wrapper_cls(model_obj=model_obj)

    @property
    def create_consent_options(self):
        """Returns a dictionary of options to create a new
        unpersisted consent model instance.
        """
        options = dict(
            subject_identifier=self.object.subject_identifier,
            consent_identifier=get_uuid(),
            version=self.consent_object.version,
        )
        return options

    @property
    def consent_options(self):
        """Returns a dictionary of options to get an existing
        consent model instance.
        """
        options = dict(
            subject_identifier=is_subject_identifier_or_raise(
                self.object.subject_identifier, reference_obj=self.object
            ),
            version=self.consent_object.version,
        )
        return options



# Generated by Django 4.2.7 on 2023-12-04 22:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_consent", "0003_edcpermissions_locale_created_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="edcpermissions",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Edc Permissions",
                "verbose_name_plural": "Edc Permissions",
            },
        ),
        migrations.AddIndex(
            model_name="edcpermissions",
            index=models.Index(
                fields=["modified", "created"], name="edc_consent_modifie_851b6a_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="edcpermissions",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_consent_user_mo_27c64b_idx",
            ),
        ),
    ]



from datetime import timedelta

from dateutil.relativedelta import relativedelta
from django.contrib.sites.models import Site
from django.test import TestCase, override_settings
from edc_protocol import Protocol
from edc_utils import get_utcnow
from model_bakery import baker

from edc_consent.field_mixins import IdentityFieldsMixinError
from edc_consent.site_consents import site_consents

from ..consent_test_utils import consent_factory
from ..models import SubjectConsent


@override_settings(
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=get_utcnow() - relativedelta(years=5),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=get_utcnow() + relativedelta(years=1),
    EDC_AUTH_SKIP_SITE_AUTHS=True,
    EDC_AUTH_SKIP_AUTH_UPDATER=False,
)
class TestConsentModel(TestCase):
    def setUp(self):
        self.study_open_datetime = Protocol().study_open_datetime
        self.study_close_datetime = Protocol().study_close_datetime
        site_consents.registry = {}
        consent_factory(
            start=self.study_open_datetime,
            end=self.study_open_datetime + timedelta(days=50),
            version="1.0",
        )
        consent_factory(
            start=self.study_open_datetime + timedelta(days=51),
            end=self.study_open_datetime + timedelta(days=100),
            version="2.0",
        )
        consent_factory(
            start=self.study_open_datetime + timedelta(days=101),
            end=self.study_open_datetime + timedelta(days=150),
            version="3.0",
            updates_versions="1.0, 2.0",
        )
        self.dob = self.study_open_datetime - relativedelta(years=25)

    def test_encryption(self):
        subject_consent = baker.make_recipe(
            "edc_consent.subjectconsent",
            first_name="ERIK",
            consent_datetime=self.study_open_datetime,
            dob=get_utcnow() + relativedelta(years=25),
        )
        self.assertEqual(subject_consent.first_name, "ERIK")

    def test_gets_subject_identifier(self):
        """Asserts a blank subject identifier is set to the
        subject_identifier_as_pk.
        """
        consent = baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=None,
            consent_datetime=self.study_open_datetime,
            dob=get_utcnow() + relativedelta(years=25),
            site=Site.objects.get_current(),
        )
        self.assertIsNotNone(consent.subject_identifier)
        self.assertNotEqual(consent.subject_identifier, consent.subject_identifier_as_pk)
        consent.save()
        self.assertIsNotNone(consent.subject_identifier)
        self.assertNotEqual(consent.subject_identifier, consent.subject_identifier_as_pk)

    def test_subject_has_current_consent(self):
        subject_identifier = "123456789"
        identity = "987654321"
        baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=self.study_open_datetime + timedelta(days=1),
            dob=get_utcnow() + relativedelta(years=25),
        )
        subject_consent = SubjectConsent.consent.consent_for_period(
            "123456789", self.study_open_datetime + timedelta(days=1)
        )
        self.assertEqual(subject_consent.version, "1.0")
        baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=self.study_open_datetime + timedelta(days=60),
            dob=get_utcnow() + relativedelta(years=25),
        )
        subject_consent = SubjectConsent.consent.consent_for_period(
            "123456789", self.study_open_datetime + timedelta(days=60)
        )
        self.assertEqual(subject_consent.version, "2.0")

    def test_model_updates(self):
        subject_identifier = "123456789"
        identity = "987654321"
        consent = baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=self.study_open_datetime,
            dob=get_utcnow() + relativedelta(years=25),
        )
        self.assertEqual(consent.version, "1.0")
        consent = baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=self.study_open_datetime + timedelta(days=51),
            dob=get_utcnow() + relativedelta(years=25),
        )
        self.assertEqual(consent.version, "2.0")
        consent = baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=self.study_open_datetime + timedelta(days=101),
            dob=get_utcnow() + relativedelta(years=25),
        )
        self.assertEqual(consent.version, "3.0")

    def test_model_updates2(self):
        subject_identifier = "123456789"
        identity = "987654321"
        consent = baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=self.study_open_datetime,
            dob=get_utcnow() + relativedelta(years=25),
        )
        self.assertEqual(consent.version, "1.0")
        consent = baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=self.study_open_datetime + timedelta(days=101),
            dob=get_utcnow() + relativedelta(years=25),
        )
        self.assertEqual(consent.version, "3.0")

    def test_manager(self):
        for i in range(1, 3):
            baker.make_recipe(
                "edc_consent.subjectconsent",
                subject_identifier=str(i),
                consent_datetime=self.study_open_datetime + relativedelta(days=i),
            )

        first = SubjectConsent.objects.get(subject_identifier="1")
        self.assertEqual(first, SubjectConsent.consent.first_consent(subject_identifier="1"))

        SubjectConsent.consent.consent_for_period(
            subject_identifier="1",
            report_datetime=self.study_open_datetime + relativedelta(days=1),
        )
        self.assertEqual(first, SubjectConsent.consent.first_consent(subject_identifier="1"))

        self.assertIsNone(
            SubjectConsent.consent.consent_for_period(
                subject_identifier="A",
                report_datetime=self.study_open_datetime - relativedelta(days=1),
            )
        )

        self.assertIsNone(
            SubjectConsent.consent.consent_for_period(
                subject_identifier="A",
                report_datetime=self.study_open_datetime + relativedelta(days=1),
            )
        )

    def test_model_str_repr_etc(self):
        obj = baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier="12345",
            consent_datetime=self.study_open_datetime + relativedelta(days=1),
        )
        self.assertTrue(str(obj))
        self.assertTrue(repr(obj))
        self.assertTrue(obj.age_at_consent)
        self.assertTrue(obj.formatted_age_at_consent)
        self.assertEqual(obj.report_datetime, obj.consent_datetime)

        self.assertRaises(
            IdentityFieldsMixinError,
            baker.make_recipe,
            "edc_consent.subjectconsent",
            subject_identifier="12345",
            consent_datetime=self.study_open_datetime + relativedelta(days=1),
            identity="123456789",
            confirm_identity="987654321",
        )



from dateutil.relativedelta import relativedelta
from django.contrib.sites.models import Site
from edc_constants.constants import MALE, NO, YES
from edc_utils import get_utcnow
from faker import Faker
from model_bakery.recipe import Recipe, seq

from .tests.models import SubjectConsent

fake = Faker()

subjectconsent = Recipe(
    SubjectConsent,
    consent_datetime=get_utcnow,
    dob=get_utcnow() - relativedelta(years=25),
    first_name=fake.first_name,
    last_name=fake.last_name,
    # note, passes for model but won't pass validation in modelform clean()
    initials="AA",
    gender=MALE,
    # will raise IntegrityError if multiple made without _quantity
    identity=seq("12315678"),
    # will raise IntegrityError if multiple made without _quantity
    confirm_identity=seq("12315678"),
    identity_type="passport",
    is_dob_estimated="-",
    language="en",
    is_literate=YES,
    is_incarcerated=NO,
    study_questions=YES,
    consent_reviewed=YES,
    consent_copy=YES,
    assessment_score=YES,
    consent_signature=YES,
    site=Site.objects.get_current(),
)



from importlib.metadata import version

__version__ = version("edc_consent")

from .exceptions import ConsentObjectDoesNotExist, NotConsentedError
from .model_wrappers import ConsentModelWrapperMixin
from .site_consents import site_consents



from django.conf import settings
from django.db import models
from django.utils.html import format_html
from django_crypto_fields.fields import LastnameField
from django_crypto_fields.models import CryptoMixin
from edc_constants.choices import YES_NO

from ..validators import FullNameValidator, eligible_if_no


class VulnerabilityFieldsMixin(CryptoMixin, models.Model):
    is_incarcerated = models.CharField(
        verbose_name="Is the participant under involuntary incarceration?",
        max_length=3,
        choices=YES_NO,
        validators=[eligible_if_no],
        null=True,
        help_text="( if 'Yes' STOP participant cannot be consented )",
    )

    is_literate = models.CharField(
        verbose_name="Is the participant literate?",
        max_length=3,
        choices=YES_NO,
        help_text=(
            "If 'No' provide witness's name on this "
            "form and signature on the paper document."
        ),
    )

    witness_name = LastnameField(
        verbose_name="Witness's last and first name",
        validators=[FullNameValidator()],
        blank=True,
        null=True,
        help_text=format_html(
            "Required only if participant is illiterate.<br>"
            "Format is 'LASTNAME, FIRSTNAME'. "
            "All uppercase separated by a comma."
        ),
    )

    language = models.CharField(
        verbose_name="Language of consent",
        max_length=25,
        choices=settings.LANGUAGES,
        help_text=(
            "The language used for the consent process will "
            "also be used during data collection."
        ),
    )

    class Meta:
        abstract = True



from edc_utils import get_uuid

from .exceptions import ConsentObjectDoesNotExist
from .site_consents import site_consents


class ConsentViewMixin:

    """Declare with edc_appointment view mixin to get `appointment`."""

    consent_model_wrapper_cls = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._consent = None
        self._consents = None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            consent=self.consent_wrapped,
            consents=self.consents_wrapped,
            consent_object=self.consent_object,
        )
        return context

    @property
    def consent_object(self):
        """Returns a consent_config object or None
        from site_consents for the current reporting period.
        """
        try:
            consent_object = site_consents.get_consent_for_period(
                model=self.consent_model_wrapper_cls.model,
                report_datetime=self.report_datetime,
            )
        except ConsentObjectDoesNotExist:
            consent_object = None
        return consent_object

    @property
    def consent(self):
        """Returns a consent model instance or None for the current period."""
        return self.consent_object.model_cls.consent.consent_for_period(
            subject_identifier=self.subject_identifier,
            report_datetime=self.report_datetime,
        )

    @property
    def consent_wrapped(self):
        """Returns a wrapped consent, either saved or not,
        for the current period.
        """
        return self.consent_model_wrapper_cls(self.consent or self.empty_consent)

    @property
    def empty_consent(self):
        """Returns an unsaved consent model instance.

        Override to include additional attrs to instantiate.
        """
        return self.consent_object.model_cls(
            subject_identifier=self.subject_identifier,
            consent_identifier=get_uuid(),
            version=self.consent_object.version,
        )

    @property
    def consents(self):
        """Returns a Queryset of consents for this subject."""
        return self.consent_object.model_cls.objects.filter(
            subject_identifier=self.subject_identifier
        ).order_by("version")

    @property
    def consents_wrapped(self):
        """Returns a list of wrapped consents that this user
        has permissions to access.
        """
        consents_wrapped = []
        for obj in self.consents:
            perm_list = []
            for code in ["add", "change", "view"]:
                perm_list.append(
                    f'{obj._meta.app_label}.{code}_{obj._meta.label_lower.split(".")[1]}'
                )
            if self.request.user.has_perms(perm_list):
                consents_wrapped.append(self.consent_model_wrapper_cls(obj))
        return consents_wrapped
